/*
 "NiseRingo" is a clone of "Niseharuna" works on MacOS X.
 Copyright (C) 2001-2003 phonohawkput

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
package aya;
import aya.*;
import aya.node.*;
import java.util.*;

public class Parser
{
	Aya aya;
	LexicalAnalyzer lex;
	
	public Parser(Aya aya,LexicalAnalyzer lex) {
		this.aya = aya;
		this.lex = lex;
	}
	
	public Function parseFunction() throws Exception
	{
		// 関数 := 関数名 ( ':' 関数オプション ) ブロック
		Token t;
		
		String name;
		t = lex.nextToken();
		if (t.getType() == LexicalAnalyzer.TYPE_IDENTIFIER) {
			name = t.getToken();
		}
		else {
			lex.rollback();
			return null;
		}
		
		t = lex.nextToken();
		String option = null;
		if (t.getType() == LexicalAnalyzer.TYPE_SYMBOL && t.getToken().equals(":")) {
			Token t_option = lex.nextToken();
			if (t_option.getType() == LexicalAnalyzer.TYPE_IDENTIFIER) {
				option = t_option.getToken();
			}
			else {
				// 識別子でなかった。文法エラー。
				throw new IllegalTokenException(t);
			}
		}
		else {
			lex.rollback();
		}
		
		Block b = parseBlock();
		if (b == null) {
			// 次に来たのがブロックでなかった。文法エラー。
			throw new IllegalTokenException(lex.nextToken());
		}
		b.setOption(option);
		
		return new Function(aya,name,b);
	}
	
	public Block parseBlock() throws Exception
	{
		// ブロック := '{' (文 ';'?)* '}'
		Token t;
		
		t = lex.nextToken();
		if (t.getType() != LexicalAnalyzer.TYPE_SYMBOL || !t.getToken().equals("{")) {
			// これはブロックでない。
			lex.rollback();
			return null;
		}
		
		// ブロックの終わりまでループ
		Block b = new Block(aya);
		while (true) {
			Statement s = parseStatement();
			if (s == null) {
				// 次に来たのが文でなかった。
				t = lex.nextToken();
				if (t.getType() == LexicalAnalyzer.TYPE_SYMBOL && t.getToken().equals(";")) {
					// セミコロンは単に無視する。
					continue;
				}
				else {
					lex.rollback();
				}
				
				t = lex.nextToken();
				if (t.getType() == LexicalAnalyzer.TYPE_SYMBOL && t.getToken().equals("}")) {
					// ブロックの終わり。
					break;
				}
				else {
					// セミコロンでもブロックの終わりでもない。文法エラー。
					throw new IllegalTokenException(t);
				}
			}
			else {
				b.add(s);
			}
		}
		
		return b;
	}
	
	public Statement parseStatement() throws Exception
	{
		// 文 := 代入文 | 式 | ブロック | '--' | if文 | case文 | switch文 | while文 | for文 | 'break' | 'continue' | 'return'
		Token t;
		
		Substitution subs = parseSubstitution();
		if (subs != null) {
			return Statement.getFactory().newStatement(subs);
		}
		
		Expression expr = parseExpression();
		if (expr != null) {
			return Statement.getFactory().newStatement(expr);
		}
		
		Block block = parseBlock();
		if (block != null) {
			return Statement.getFactory().newStatement(block);
		}
		
		Statement.If s_if = parseIf();
		if (s_if != null) {
			return s_if;
		}
		
		Case s_case = parseCase();
		if (s_case != null) {
			return s_case;
		}
		
		Statement.Switch s_switch = parseSwitch();
		if (s_switch != null) {
			return s_switch;
		}
		
		Statement.While s_while = parseWhile();
		if (s_while != null) {
			return s_while;
		}
		
		Statement.For s_for = parseFor();
		if (s_for != null) {
			return s_for;
		}
		
		t = lex.nextToken();
		if (t.getType() == LexicalAnalyzer.TYPE_SYMBOL) {
			if (t.getToken().equals("--")) {
				return Statement.getFactory().newSeparate();
			}
			else if (t.getToken().equals("break")) {
				return Statement.getFactory().newBreak();
			}
			else if (t.getToken().equals("continue")) {
				return Statement.getFactory().newContinue();
			}
			else if (t.getToken().equals("return")) {
				return Statement.getFactory().newReturn();
			}
			else {
				// これは文でない。戻る。
				lex.rollback();
				return null;
			}
		}
		else {
			lex.rollback();
			return null;
		}
	}
	
	public Expression parseExpression() throws Exception
	{
		// 式 := 項 (符号 項)*
		Expression expr = new Expression(aya);
		boolean firsttime = true;
		while (true) {
			int signal = 1;
			if (firsttime) {
				firsttime = false;
			}
			else {
				Token t = lex.nextToken();
				if (t.isSymbol()) {
					if (t.getToken().equals("+")) {
						// この項は正の符号を持つ
					}
					else if (t.getToken().equals("-")) {
						signal = -1;
					}
					else {
						lex.rollback();
						break;
					}
				}
				else {
					lex.rollback();
					break;
				}
			}
			
			Expression.Term tm = parseTerm();
			if (tm == null) {
				// 項でなかった。
				break;
			}
			else {
				tm.setSignal(signal);
				expr.addTerm(tm);
			}
		}
		
		if (expr.size() == 0) {
			// 項が一つも無いので、これは式でない。
			return null;
		}
		else {
			return expr;
		}
	}
	
	public Expression.Term parseTerm() throws Exception
	{
		// 項 := 符号? 因子 (因子結合子 符号? 因子)*
		// 符号 := '+' | '-'
		// 因子結合子 := '*' | '/' | '%'
		Token t;
		Expression.Term term = Expression.getFactory().newTerm();
		
		boolean firsttime = true;
		while (true) {
			char combination = '?';
			// 最初だけは因子結合子を見ない。
			if (firsttime) {
				firsttime = false;
			}
			else {
				t = lex.nextToken();
				if (t.getType() == LexicalAnalyzer.TYPE_SYMBOL) {
					if (t.getToken().equals("*")) {
						combination = '*';
					}
					else if (t.getToken().equals("/")) {
						combination = '/';
					}
					else if (t.getToken().equals("%")) {
						combination = '%';
					}
					else {
						// 因子結合子でないので、ここで終わり。
						lex.rollback();
						break;
					}
				}
				else {
					// 因子結合子でないので、ここで終わり。
					lex.rollback();
					break;
				}
			}
		
			t = lex.nextToken();
			int signal = 1;
			if (t.getType() == LexicalAnalyzer.TYPE_SYMBOL) {
				if (t.getToken().equals("+")) {
					// 符号+
				}
				else if (t.getToken().equals("-")) {
					// 符号-
					signal = -1;
				}
				else {
					// どちらでもない。戻る。
					lex.rollback();
				}
			}
			else {
				// シンボルでない。戻る。
				lex.rollback();
			}
			
			Expression.Factor factor = parseFactor();
			if (factor == null) {
				// 因子でないので、これは項でない。
				return null;
			}
			factor.setSignal(signal);
			
			term.addFactor(combination,factor);
		}
		return term;
	}
		
	public Expression.Factor parseFactor() throws Exception
	{
		// 因子 := '(' 式 ')' | <数値定数> | <文字列定数> | 関数呼出し | 変数
		Token t;
		
		FunctionCall fcall = parseFunctionCall();
		if (fcall != null) {
			return Expression.getFactory().newFactor(aya,fcall);
		}
		
		VariableReference vref = parseVariable();
		if (vref != null) {
			return Expression.getFactory().newFactor(aya,vref);
		}
		
		t = lex.nextToken();
		if (t.getType() == LexicalAnalyzer.TYPE_SYMBOL) {
			if (t.getToken().equals("(")) {
				// 次に来るのは式であるはず。
				Expression expr = parseExpression();
				if (expr == null) {
					// 文法エラー
					throw new IllegalTokenException(t);
				}
				
				// 次に来るのは閉じ括弧であるはず。
				t = lex.nextToken();
				if (t.getType() == LexicalAnalyzer.TYPE_SYMBOL && t.getToken().equals(")")) {
					return Expression.getFactory().newFactor(expr);
				}
				else {
					// 文法エラー。
					throw new IllegalTokenException(t);
				}
			}
			else {
				// これは因子でない。戻る。
				lex.rollback();
				return null;
			}
		}
		else if (t.getType() == LexicalAnalyzer.TYPE_STRING) {
			return Expression.getFactory().newFactor(aya,t.getToken());
		}
		else if (t.getType() == LexicalAnalyzer.TYPE_NUMBER) {
			return Expression.getFactory().newFactor(parseNumericConstant(t.getToken()));
		}
		else {
			// 因子でない。
			lex.rollback();
			return null;
		}
	}
	
	public static double parseNumericConstant(String num) throws Exception {
		double n;
		if (num.startsWith("0x")) {
			n = Long.parseLong(num.substring(2),16);
		}
		else if (num.startsWith("0b")) {
			n = Long.parseLong(num.substring(2),2);
		}
		else {
			n = Double.parseDouble(num);
		}
		return n;
	}
	
	public FunctionCall parseFunctionCall() throws Exception
	{
		// 関数呼出し := 関数名 引数群?
		// 引数群 := '(' (引数? | 引数 (',' 引数)*) ')'
		
		// 引数群が無く、その代わりにトークン'['が存在する場合は
		// 添字付き変数参照なので、関数呼出しとは見做さない。
		Token t;
		
		Token t_name = t = lex.nextToken();
		String name;
		if (t.getType() == LexicalAnalyzer.TYPE_IDENTIFIER) {
			name = t.getToken();
		}
		else {
			// 識別子でないので、これは関数呼出しではない。
			lex.rollback();
			return null;
		}
		
		FunctionCall fcall = new FunctionCall(name);
		t = lex.nextToken();
		if (t.getType() == LexicalAnalyzer.TYPE_SYMBOL && t.getToken().equals("(")) {
			// 開き括弧。引数群がある。
		}
		else {
			// 開き括弧でないので、引数群が無い。
			// 次に来るのが'['か？
			if (t.getType() == LexicalAnalyzer.TYPE_SYMBOL && t.getToken().equals("[")) {
				// そもそも関数呼出しでない。
				lex.rollback();
				lex.rollback();
				return null;
			}
			else {
				lex.rollback();
				return fcall;
			}
		}
		
		fcall.prepareArgs();
		boolean firsttime = true;
		while (true) {
			// 最初の引数でなければカンマを見る。
			if (firsttime) {
				firsttime = false;
				
				t = lex.nextToken();
				if (t.isSymbol() && t.getToken().equals(")")) {
					// 引数無し。ここで終わり。
					break;
				}
				else {
					lex.rollback();
				}
			}
			else {
				t = lex.nextToken();
				if (t.getType() == LexicalAnalyzer.TYPE_SYMBOL && t.getToken().equals(",")) {
					// 次の引数へ
				}
				else if (t.getType() == LexicalAnalyzer.TYPE_SYMBOL && t.getToken().equals(")")) {
					break; // ここで終わり
				}
				else {
					// カンマでも閉じ括弧でもないので、これは関数呼出しではない。文法エラー。
					throw new IllegalTokenException(t);
				}
			}
			
			FunctionCall.Argument arg = parseArgument();
			if (arg == null) {
				// 引数だと思っていたのに、引数でなかった。文法エラー。
				throw new IllegalTokenException(lex.nextToken());
			}
			else {
				fcall.addArg(arg);
			}
		}
		
		return fcall;
	}
	
	public FunctionCall.Argument parseArgument() throws Exception
	{
		// 引数 := 式 | '&' 変数
		Token t;
		
		Expression expr = parseExpression();
		if (expr != null) {
			return FunctionCall.getFactory().newArg(expr);
		}
		
		t = lex.nextToken();
		if (t.getType() == LexicalAnalyzer.TYPE_SYMBOL && t.getToken().equals("&")) {
			// 次に来るのは変数か?
			VariableReference vref = parseVariable();
			if (vref == null) {
				// 変数でないので、これは引数でない。文法エラー。
				throw new IllegalTokenException(lex.nextToken());
			}
			else {
				return FunctionCall.getFactory().newArg(vref);
			}
		}
		else {
			// これは引数でない。
			lex.rollback();
			return null;
		}
	}
	
	public VariableReference parseVariable() throws Exception
	{
		// 変数 := 変数名 ('[' 式 ']')?
		Token t;
		
		t = lex.nextToken();
		String name;
		if (t.getType() == LexicalAnalyzer.TYPE_IDENTIFIER) {
			name = t.getToken();
		}
		else {
			// これは変数でない。
			lex.rollback();
			return null;
		}
		VariableReference vref = new VariableReference(name);
		
		t = lex.nextToken();
		if (t.getType() == LexicalAnalyzer.TYPE_SYMBOL && t.getToken().equals("[")) {
			// 添字がある。
			Expression expr = parseExpression();
			if (expr == null) {
				// 中身が式でなかった。これは変数参照でない。
				return null;
			}
			else {
				vref.setIndex(expr);
			}
			
			t = lex.nextToken();
			if (t.getType() == LexicalAnalyzer.TYPE_SYMBOL && t.getToken().equals("]")) {
				// 正しい添字の終わり
			}
			else {
				// 正しく添字が終わっていない。文法エラー。
				throw new IllegalTokenException(t);
			}
		}
		else {
			// 添字が無い。
			lex.rollback();
		}
		
		return vref;
	}
	
	public Substitution parseSubstitution() throws Exception
	{
		// 代入文 := 変数 インクリメント演算子 | 変数 代入演算子 式
		// インクリメント演算子 := '++' | '--'
		// 代入演算子 := '='  | '+='  | '-='  | '*='  | '/='  | '%=' |
		//				 ':=' | '+:=' | '-:=' | '*:=' | '/:=' | '%:='
		lex.push();
		VariableReference vref = parseVariable();
		if (vref == null) {
			// 変数でないので、これは代入文でない。
			lex.popnback();
			return null;
		}
		Substitution subs = new Substitution(aya,vref);
		
		Token t;
		t = lex.nextToken();
		if (t.getType() == LexicalAnalyzer.TYPE_SYMBOL) {
			String oper = t.getToken();
			if (oper.equals("++") || oper.equals("--")) {
				// インクリメント
				if (oper.equals("--") && t.wasSpaceBefore()) {
					// --の前に空白文字があった場合、これは出力確定子の--である。
					// これは代入文でない。
					lex.popnback();
					return null;
				}
				subs.setIncrementalOper(oper);
			}
			else if (oper.equals("=") || oper.equals("+=") || oper.equals("-=") ||
					 oper.equals("*=") || oper.equals("/=") || oper.equals("%=") ||
					 oper.equals(":=") || oper.equals("+:=") || oper.equals("-:=") ||
					 oper.equals("*:=") || oper.equals("/:=") || oper.equals("%:=")) {
				// 代入
				// 次に来た式を得る。
				Expression expr = parseExpression();
				if (expr == null) {
					// 文法エラー。
					throw new IllegalTokenException(lex.nextToken());
				}
				subs.setSubstitutionalOper(oper,expr);
			}
			else {
				// 代入文で用いられる演算子でない。これは代入文でない。
				lex.popnback();
				return null;
			}
		}
		else {
			// 次に来たのが演算子でない。これは代入文でない。
			lex.popnback();
			return null;
		}
		
		lex.pop();
		return subs;
	}
	
	public Statement.If parseIf() throws Exception
	{
		// if文 := 'if' 条件式 ブロック (elseif 条件式 ブロック)* (else ブロック)?
		Token t;
		
		t = lex.nextToken();
		if (t.getType() != LexicalAnalyzer.TYPE_SYMBOL || !t.getToken().equals("if")) {
			// if文でない。
			lex.rollback();
			return null;
		}
		
		Condition head_cond = parseCondition();
		if (head_cond == null) {
			// ifの次に来たのが条件式でない。文法エラー。
			throw new IllegalTokenException(lex.nextToken());
		}
		
		Block head_block = parseBlock();
		if (head_block == null) {
			// 次に来たのがブロックでない。文法エラー。
			throw new IllegalTokenException(lex.nextToken());
		}
		
		Statement.If stat_if = Statement.getFactory().newIf(head_cond,head_block);
		
		while (true) {
			t = lex.nextToken();
			if (t.getType() == LexicalAnalyzer.TYPE_SYMBOL && t.getToken().equals("elseif")) {
				
				Condition cond = parseCondition();
				if (cond == null) {
					// elseifの次に来たのが条件式でない。文法エラー。
					throw new IllegalTokenException(lex.nextToken());
				}
				
				Block block = parseBlock();
				if (block == null) {
					// 次に来たのがブロックでない。文法エラー。
					throw new IllegalTokenException(lex.nextToken());
				}
				
				stat_if.addElseIf(cond,block);
			}
			else {
				// elseifでない。ここで終わり。
				lex.rollback();
				break;
			}
		}
		
		t = lex.nextToken();
		if (t.getType() == LexicalAnalyzer.TYPE_SYMBOL && t.getToken().equals("else")) {
			
			Block block = parseBlock();
			if (block == null) {
				// 次に来たのがブロックでない。文法エラー。
				throw new IllegalTokenException(lex.nextToken());
			}
			
			stat_if.setElse(block);
		}
		else {
			// elseでない。
			lex.rollback();
		}
		
		return stat_if;
	}
	
	public Condition parseCondition() throws Exception
	{
		// 条件式 := 部分条件式 (条件結合子 部分条件式)*
		// 条件結合子 := '||' | '&&'
		Token t;
		
		Condition.Subcondition head = parseSubcondition();
		if (head == null) {
			// 条件式でない。
			return null;
		}
		Condition cond = new Condition(head);
		
		// 次に条件結合子が来ているか？
		while (true) {
			t = lex.nextToken();
			if (t.getType() == LexicalAnalyzer.TYPE_SYMBOL &&
				(t.getToken().equals("||") || t.getToken().equals("&&"))) {
				
				// これは条件結合子である。次に来るのは部分条件式でなければならない。
				Condition.Subcondition sub = parseSubcondition();
				if (sub == null) {
					// 文法エラー。
					throw new IllegalTokenException(lex.nextToken());
				}
				
				cond.add(t.getToken(),sub);
			}
			else {
				// 条件結合子でない。ここで条件式終わり。
				lex.rollback();
				break;
			}
		}
		
		return cond;
	}
	
	public Condition.Subcondition parseSubcondition() throws Exception
	{
		// 部分条件式 := '(' 条件式 ')' | 式 条件比較子 式
		// 条件比較子 := '==' | '!=' | '<' | '<=' | '>' | '>=' | '_in_' | '!_in_'
		Token t;
		
		// '(' 条件式 ')'という形式は一見すると式のように見えるため(実際には違うのでエラー)
		// 先にparseExpressionを実行するわけにはいかない。
		
		t = lex.nextToken();
		if (t.getType() == LexicalAnalyzer.TYPE_SYMBOL && t.getToken().equals("(")) {
			// 条件式かも知れないので条件式としてパースしてみる。
			Condition cond = parseCondition();
			if (cond == null) {
				// 条件式ではなかった。戻る。
				lex.rollback();
			}
			else {
				// 条件式だ。次に来たのは')'か？
				t = lex.nextToken();
				if (t.getType() == LexicalAnalyzer.TYPE_SYMBOL && t.getToken().equals(")")) {
					// 閉じ括弧だった。
					return Condition.getFactory().newSubcondition(cond);
				}
				else {
					// 違う。文法エラー。
					throw new IllegalTokenException(t);
				}
			}
		}
		else {
			// 括弧でなかった。戻る。
			lex.rollback();
		}
		
		// '(' 条件式 ')'でなかった場合のみ此処に来る。
		lex.push();
		Expression left = parseExpression();
		if (left == null) {
			// 部分条件式でない。
			lex.popnback();
			return null;
		}
		
		String cmp;
		Expression right;
		
		t = lex.nextToken();
		if (t.getType() == LexicalAnalyzer.TYPE_SYMBOL) {
			cmp = t.getToken();
			
			// 誤り訂正！
			if (cmp.equals("=")) {
				cmp = "==";
			}
			
			if (cmp.equals("==") || cmp.equals("!=") || cmp.equals("<") || cmp.equals("<=") ||
				cmp.equals(">") || cmp.equals(">=") || cmp.equals("_in_") || cmp.equals("!_in_")) {
				// これは条件比較子である。次に来るのは条件因子でなければならない。
				right = parseExpression();
				if (right == null) {
					// 文法エラー
					throw new IllegalTokenException(lex.nextToken());
				}
			}
			else {
				// 次に来たのが条件比較子でないので、これは部分条件式でない。
				lex.popnback();
				return null;
			}
		}
		else {
			// 次に来たのが条件比較子でないので、これは部分条件式でない。
			lex.popnback();
			return null;
		}
		
		lex.pop();
		return Condition.getFactory().newSubcondition(left,cmp,right);
	}
	
	public Case parseCase() throws Exception
	{
		// case文 := 'case' 式 '{' case候補+ '}'
		// case候補 := 'when' case条件 ブロック | 'others' ブロック
		Token t;
		
		t = lex.nextToken();
		if (!t.isSymbol() || !t.getToken().equals("case")) {
			// caseでない。
			lex.rollback();
			return null;
		}
		
		Expression attention = parseExpression();
		if (attention == null) {
			// caseの次に来たのが式でない。文法エラー。
			throw new IllegalTokenException(lex.nextToken());
		}
		Case stat_case = new Case(attention);
		
		t = lex.nextToken();
		if (!t.isSymbol() || !t.getToken().equals("{")) {
			// 式の次に来たのが'{'でない。文法エラー。
			throw new IllegalTokenException(t);
		}
		
		while (true) {
			t = lex.nextToken();
			
			if (t.isSymbol() && t.getToken().equals("when")) {
				// 次に来るのはcase条件であるはず。
				Case.CaseCondition cond = parseConditionOfCase();
				if (cond == null) {
					throw new IllegalTokenException(lex.nextToken());
				}
				
				// その次に来るのはブロックであるはず。
				Block block = parseBlock();
				if (block == null) {
					throw new IllegalTokenException(lex.nextToken());
				}
				
				stat_case.addCandidate(stat_case.newCandidate(cond,block));
			}
			else if (t.isSymbol() && t.getToken().equals("others")) {
				// 次に来るのはブロックであるはず。
				Block block = parseBlock();
				if (block == null) {
					throw new IllegalTokenException(lex.nextToken());
				}
				
				stat_case.addCandidate(stat_case.newCandidate(null,block));
			}
			else {
				// ここで終わり
				lex.rollback();
				break;
			}
		}
		
		t = lex.nextToken();
		if (!t.isSymbol() || !t.getToken().equals("}")) {
			// 次に来たのが'}'でない。文法エラー。
			throw new IllegalTokenException(t);
		}
		
		return stat_case;
	}
	
	public Case.CaseCondition parseConditionOfCase() throws Exception
	{
		// case条件 := case部分条件 (',' case部分条件)*
		// case部分条件 := <数値定数> | <数値定数> '-' <数値定数> | <文字列定数> | <文字列定数> '-' <文字列定数>
		Token t;
		
		Case.CaseCondition cond = Case.getCaseFactory().newCondition();
		boolean firsttime = true;
		while (true) {
			if (firsttime) {
				firsttime = false;
			}
			else {
				t = lex.nextToken();
				if (!t.isSymbol() || !t.getToken().equals(",")) {
					// カンマでない。ここで終わり。
					lex.rollback();
					break;
				}
			}
		
			t = lex.nextToken();
			
			if (t.isNumber()) {
				// '-' 数値 と続いていれば、数値範囲。
				// そうでなければ数値定数。
				String lower = t.getToken();
				
				t = lex.nextToken();
				if (t.isSymbol() && t.getToken().equals("-")) {
					t = lex.nextToken();
					if (t.isNumber()) {
						String higher = t.getToken();
						
						cond.addSubcondition(
							Case.getCaseFactory().newSubcondition(
								parseNumericConstant(lower),parseNumericConstant(higher)));
					}
					else {
						throw new IllegalTokenException(t);
					}
				}
				else {
					lex.rollback();
					cond.addSubcondition(
						Case.getCaseFactory().newSubcondition(parseNumericConstant(lower)));
				}
			}
			else if (t.isString()) {
				// '-' 文字列定数 と続いていれば、文字列範囲。
				// そうでなければ数値定数。
				String lower = t.getToken();
				
				t = lex.nextToken();
				if (t.isSymbol() && t.getToken().equals("-")) {
					t = lex.nextToken();
					if (t.isString()) {
						String higher = t.getToken();
						
						cond.addSubcondition(
							Case.getCaseFactory().newSubcondition(lower,higher));
					}
					else {
						throw new IllegalTokenException(t);
					}
				}
				else {
					lex.rollback();
					cond.addSubcondition(
						Case.getCaseFactory().newSubcondition(lower));
				}
			}
			else {
				throw new IllegalTokenException(t);
			}
		}
		
		return cond;
	}
	
	public Statement.Switch parseSwitch() throws Exception
	{
		// switch文 := 'switch' 式 ブロック
		Token t;
		
		t = lex.nextToken();
		if (!t.isSymbol() || !t.getToken().equals("switch")) {
			// これはswitch文でない。
			lex.rollback();
			return null;
		}
		
		Expression expr = parseExpression();
		if (expr == null) {
			throw new IllegalTokenException(lex.nextToken());
		}
		
		Block block = parseBlock();
		if (block == null) {
			throw new IllegalTokenException(lex.nextToken());
		}
		
		return Statement.getFactory().newSwitch(expr,block);
	}
	
	public Statement.While parseWhile() throws Exception
	{
		// while文 := 'while' 条件式 ブロック
		Token t;
		
		t = lex.nextToken();
		if (!t.isSymbol() || !t.getToken().equals("while")) {
			// これはwhile文でない。
			lex.rollback();
			return null;
		}
		
		Condition cond = parseCondition();
		if (cond == null) {
			throw new IllegalTokenException(lex.nextToken());
		}
		
		Block block = parseBlock();
		if (block == null) {
			throw new IllegalTokenException(lex.nextToken());
		}
		
		return Statement.getFactory().newWhile(cond,block);
	}
	
	public Statement.For parseFor() throws Exception
	{
		// for文 := 'for' 代入文? ';' 条件式? ';' 代入文? ブロック
		Token t;
		
		t = lex.nextToken();
		if (!t.isSymbol() || !t.getToken().equals("for")) {
			// これはwhile文でない。
			lex.rollback();
			return null;
		}
		
		Substitution init = parseSubstitution();
		
		t = lex.nextToken();
		if (!t.isSymbol() || !t.getToken().equals(";")) {
			throw new IllegalTokenException(t);
		}
		
		Condition cond = parseCondition();
		
		t = lex.nextToken();
		if (!t.isSymbol() || !t.getToken().equals(";")) {
			throw new IllegalTokenException(t);
		}
		
		Substitution alteration = parseSubstitution();
		
		Block block = parseBlock();
		if (block == null) {
			throw new IllegalTokenException(lex.nextToken());
		}
		
		return Statement.getFactory().newFor(init,cond,alteration,block);
	}
}
