/*
 "NiseRingo" is a clone of "Niseharuna" works on MacOS X.
 Copyright (C) 2001-2003 phonohawk

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
package aya.eval;
import aya.*;
import aya.node.*;
import aya.eval.*;
import java.util.*;
import java.util.regex.*;
import java.lang.reflect.*;
import java.io.*;

// このクラスはスレッドセーフではありません。
// 複数の文インタプリタが同時に走る事は出来ますが、
// 一つの文インタプリタ内で同時にシステムコールを呼ぶ事は出来ません。

public class SystemCall extends Function
{
	Method function;
	Hashtable cache; // {関数名 => SystemCall}
	
	public SystemCall(Aya aya) {
		// ファクトリを作るコンストラクタ。
		super(aya,null,null);
		this.function = null;
		this.cache = new Hashtable();
	}
	
	protected SystemCall(Aya aya,String name,Method function) {
		super(aya,name,null);
		this.function = function;
		this.cache = null;
	}
	
	static final Pattern pat_period = Pattern.compile("\\.");
	public SystemCall get(String name) {
		// 該当するシステムコールが無ければnullを返す。
		
		// ピリオドは__period__に変換する。
		if (name.indexOf('.') != -1) {
			name = pat_period.matcher(name).replaceAll("__period__");
		}
		
		SystemCall cached = (SystemCall)cache.get(name);
		if (cached != null) {
			return cached;
		}
		
		try {
			Method function = this.getClass().getMethod("syscall_"+name,new Class[] {Vector.class});
			SystemCall syscall = new SystemCall(getAya(),name,function);
			cache.put(name,syscall);
			return syscall;
		}
		catch (Exception e) {
			return null;
		}
	}
	
	public Value eval(Vector args) throws Exception {
		// args: 要素がVariablePointerまたはValueである配列。nullでも良い。
		try {
			return (Value)function.invoke(this,new Object[] {args});
		}
		catch (Exception e) {
			throw new Exception("Calling systemcall failured: "+getName()+"("+(args == null ? "" : args.toString())+")",e);
		}
	}
	
	// ---------------- 汎用 ---------------- //
	
	public Value syscall_RAND(Vector args) throws Exception {
		// RAND(n)
		// 0〜n-1 の範囲でランダム値を得ます。	
		// n は省略可能で、この場合0〜99の乱数を生成します。
		long limit;
		if (args != null && args.size() > 0) {
			limit = content(args.elementAt(0)).getInteger();
		}
		else {
			limit = 100;
		}
		
		if (limit == 0) {
			return new Value(0);
		}
		else {
			return new Value((long)(Math.random() * Long.MAX_VALUE) % limit);
		}
	}
	
	public Value syscall_ASC(Vector args) throws Exception {
		// ASC(c)
		// ASCII文字を返します。
		// 指定範囲は 0≦c＜0x80 で、範囲外の c を与えると半角空白(20H)に代替されます。
		long n = content(args.elementAt(0)).getInteger();
		if (n < 0 || n >= 0x80) {
			n = 0x20;
		}
		
		char c = (char)n;
		return new Value(""+c);
	}
	
	public Value syscall_IASC(Vector args) throws Exception {
		// IASC("foo")
		// 文字列先頭の 1 バイトの文字コードを返します。半角文字しか処理できないことに注意してください。
		// 数値や空文字列を与えると -1 が返ります。
		byte[] bytes = content(args.elementAt(0)).getString().getBytes(Aya.CHARSET);
		return new Value(bytes[0]);
	}
	
	public Value syscall_ISINSIDE(Vector args) throws Exception {
		// ISINSIDE(val,-5,10)
		// ISINSIDE(val,"foo","bar")
		// 数値および文字列が指定範囲内にあるかチェックします。
		// 引数は 3 つです。第一引数はチェック対象の値(A)。第二、三引数は範囲を指定する値(i0、i1）です。
		// i0 ≦ A ≦ i1 で 1 、それ以外で 0 が返ります。i0 と i1 の大小は気にしなくて構いません。i0 のほうが大きくても正しく動作します。
		// チェック対象の価が文字列の場合は辞書順で比較されます。
		Value attention = content(args.elementAt(0));
		if (attention.isNumeric()) {
			double n = attention.getReal();
			double x = content(args.elementAt(1)).getReal();
			double y = content(args.elementAt(2)).getReal();
			
			return new Value((x <= n && n <= y) || (y <= n && n <= x) ? 1 : 0);
		}
		else {
			String str = attention.getString();
			String x = content(args.elementAt(1)).getString();
			String y = content(args.elementAt(2)).getString();
			
			return new Value((x.compareTo(str) <= 0 && str.compareTo(y) <= 0) ||
							 (y.compareTo(str) <= 0 && str.compareTo(x) <= 0) ? 1 : 0);
		}
	}
	
	// ---------------- 型チェック ---------------- //
	
	public Value syscall_ISREAL(Vector args) throws Exception {
		// ISREAL(n)
		// 値が数値（実数）かチェックします。実数には当然ながら整数が含まれます。
		// 戻り値　0/1 = 文字列/実数。
		return new Value(content(args.elementAt(0)).isNumeric() ? 1 : 0);
	}
	
	public Value syscall_ISINTEGER(Vector args) throws Exception {
		// ISINTEGER(n)
		// 値が（小数点以下の位がない）整数かチェックします。
		// 戻り値　0/1 = 文字列 or 小数点以下の位がある数値/整数。 
		return new Value(content(args.elementAt(0)).isInteger() ? 1 : 0);
	}
	
	public Value syscall_ISFUNCTION(Vector args) throws Exception {
		// ISFUNCTION("foo")
		// 文字列が関数名かチェックします。
		// 戻り値　0/1/2 = 関数名ではない/関数名/システム関数名。
		String name = content(args.elementAt(0)).getString();
		
		Function f = getAya().getDictionary().getFunction(name);
		if (f == null) {
			return new Value(0);
		}
		else if (f instanceof SystemCall) {
			return new Value(2);
		}
		else if (f instanceof Function) {
			return new Value(1);
		}
		else {
			throw new Exception();
		}
	}
	
	// ---------------- 型変換 ---------------- //
	
	public Value syscall_TONUMBER(Vector args) throws Exception {
		// TONUMBER(var)
		// システム関数 TONUMBER は、変数に格納された文字列を数値に変換します。
		VariablePointer vptr = (VariablePointer)args.elementAt(0);
		vptr.store(vptr.fetch().setReal());
		return null;
	}
	
	public Value syscall_TOSTRING(Vector args) throws Exception {
		// TOSTRING(var)
		// TOSTRING は TONUMBER と逆の操作、すなわち数値から文字列への変換を行います。
		VariablePointer vptr = (VariablePointer)args.elementAt(0);
		vptr.store(vptr.fetch().setString());
		return null;
	}
	
	public Value syscall_TONUMBER2(Vector args) throws Exception {
		// TONUMBER2("25")
		// TONUMBER2 は TONUMBER と機能は同等ですが、結果は戻り値として出力し、引数の内容にはタッチしません。
		return ((Value)content(args.elementAt(0)).clone()).setReal();
	}
	
	public Value syscall_TOSTRING2(Vector args) throws Exception {
		// TOSTRING2(12)
		// TOSTRING2 は TOSTRING と機能は同等ですが、結果は戻り値として出力し、引数の内容にはタッチしません。
		return ((Value)content(args.elementAt(0)).clone()).setString();
	}
	
	public Value syscall_TOBINSTR(Vector args) throws Exception {
		// TOBINSTR(3)
		// TOBINSTR は 数値を 2 進数表記の文字列へ変換します。
		return new Value(Long.toBinaryString(content(args.elementAt(0)).getInteger()));
	}
	
	public Value syscall_TOHEXSTR(Vector args) throws Exception {
		// TOHEXSTR(5)
		// TOBINSTR は 数値を 16 進数表記の文字列へ変換します。
		return new Value(Long.toHexString(content(args.elementAt(0)).getInteger()));
	}
	
	public Value syscall_BINSTRTONUM(Vector args) throws Exception {
		// BINSTRTONUM("10110")
		// BINSTRTONUM は TOBINSTR の逆操作を行います。すなわち、2 進数表記の文字列を数値へ変換します。
		return new Value(Long.parseLong(content(args.elementAt(0)).getString(),2));
	}
	
	public Value syscall_HEXSTRTONUM(Vector args) throws Exception {
		// HEXSTRTONUM("2f3a")
		// HEXSTRTONUM は TOHEXSTR の逆操作を行います。すなわち、16 進数表記の文字列を数値へ変換します。
		return new Value(Long.parseLong(content(args.elementAt(0)).getString(),16));
	}
	
	// ---------------- 簡易配列関連 ---------------- //
	
	public Value syscall_ARRAYSIZE(Vector args) throws Exception {
		// ARRAYSIZE("foo,bar,baz")
		// ARRAYSIZE("foo,bar,baz",num)
		// ARRAYSIZE システム関数は、簡易配列の要素数を求めます。
		// 引数は 1 つもしくは 2 つです。第一引数は要素数を調べる文字列。第二引数は結果を取得する変数ですが、これを省略すると結果を戻り値から取得できます。
		// この関数に数値を渡してはいけません。渡した場合 0 が返ります。
		Value result = null;
		
		Object obj = args.elementAt(0);
		if (obj instanceof Value) {
			Value val = (Value)obj;
			if (val.isNumeric()) {
				result = new Value(0);
			}
			else {
				// デフォルトはカンマ区切り
				String str = val.getString();
				int n_comma = 0;
				for (int i = 0; i < str.length(); i++) {
					if (str.charAt(i) == ',') {
						n_comma++;
					}
				}
				result = new Value(n_comma + 1);
			}
		}
		else if (obj instanceof VariablePointer) {
			VariablePointer vptr = (VariablePointer)obj;
			if (vptr.getVariableReference().getIndex() != null) {
				throw new Exception("ARRAYSIZE to element of array.");
			}
			result = new Value(vptr.getVariable().getArraySize());
		}
		else {
			throw new Exception();
		}
		
		if (args.size() >= 2) {
			VariablePointer vptr = (VariablePointer)args.elementAt(1);
			vptr.store(result);
		}
		return result;
	}
	
	public Value syscall_SETSEPARATOR(Vector args) throws Exception {
		// SETSEPARATOR(foo,"-")
		// 簡易配列のセパレータは標準では半角カンマ "," ですが、SETSEPATRATOR システム関数を使用することによって、変数ごとに異なる文字列へ変更することができます。
		// ただし、システム変数のセパレータはデフォルトの半角カンマから変更することはできません。
		// 
		// 引数は 2 つで、第一引数に新しいセパレータを設定する変数、第二引数に新しいセパレータ文字列を指定します。戻り値はありません。
		VariablePointer vptr = (VariablePointer)args.elementAt(0);
		vptr.getVariable().setSeparator(content(args.elementAt(1)).getString());
		return null;
	}
	
	// ---------------- 変数/関数の間接実行 ---------------- //
	
	public Value syscall_NAMETOVALUE(Vector args) throws Exception {
		// NAMETOVALUE は、変数名を示す文字列からその内容を取得します。
		// 引数は 1 つもしくは 2 つです。引数に変数名を示す文字列のみを指定すると、戻り値でその変数の値が取得できます。第一引数に結果を取得する変数、第二引数に変数名を示す文字列を指定することもでき、この場合戻り値はありません。
		// 指定した変数が存在しなかった場合は空文字列が返ります。
		VariablePointer to_store = null;
		String varname = null;
		if (args.size() == 1) {
			varname = content(args.elementAt(0)).getString();
		}
		else if (args.size() == 2) {
			to_store = (VariablePointer)args.elementAt(0);
			varname = content(args.elementAt(1)).getString();
		}
		
		// 制限： 現状ではグローバル変数に対してのみ使用可能。
		if (varname.charAt(0) == '_') {
			throw new Exception("system-call NAMETOVALUE is currently restricted to be used with only global variable.");
		}
		
		Variable v = getAya().getGlobalNamespace().get(varname);
		Value result;
		if (v == null) {
			result = new Value("");
		}
		else {
			result = v.getValue();
		}
		
		if (to_store == null) {
			return result;
		}
		else {
			to_store.store(result);
			return null;
		}
	}
	
	public Value syscall_LETTONAME(Vector args) throws Exception {
		// LETTONAME("name", "さくら")
		// LETTONAME は NAMETOVALUE の逆操作で、変数名を示す文字列へ値を代入します。
		// 指定した変数が存在しない場合は変数が作成されます。
		// 戻り値はありません。
		String varname = content(args.elementAt(0)).getString();
		Value val = content(args.elementAt(1));
		
		// 制限： 現状ではグローバル変数に対してのみ使用可能。
		if (varname.charAt(0) == '_') {
			throw new Exception("system-call LETTONAME is currently restricted to be used with only global variable.");
		}
		
		Variable v = getAya().getGlobalNamespace().define(varname);
		v.setValue(val);
		return null;
	}
	
	public Value syscall_CALLBYNAME(Vector args) throws Exception {
		/*
			関数名から関数を実行します。
			実行した関数の戻り値が、そのまま CALLBYNAME の戻り値となります。
			
			result = CALLBYNAME("add(1,2)")
			
			上の例では加算を行う関数 add(1,2) を実行し、結果を result に代入しています。
		*/
		// Expression.StringFactorのevalを流用する。
		String str = '%' + content(args.elementAt(0)).getString();
		return Expression.getFactory().newFactor(getAya(),str).eval(getAya().getGlobalNamespace());
	}
	
	// ---------------- デバッグ ---------------- //
	
	public Value syscall_LOGGING(Vector args) throws Exception {
		// 本来は第一引数の式とその値をファイルに出力する関数だが、
		// プリプロセッサでも使わない限り実装が面倒なので、やるとしても後で。
		System.err.println("LOGGING: "+content(args.elementAt(0)).getString());
		return null;
	}
	
	public Value syscall_GETLASTERROR(Vector args) throws Exception {
		// GETLASTERROR
		// 以下の関数が失敗した場合（0 や -1 を返してきた場合。関数により異なる）、直後に GETLASTERROR を呼ぶことで失敗した理由を読み出すことができます。
		// LOADLIB , FOPEN , FCOPY , FMOVE , FDELETE , FRENAME , FSIZE , MKDIR , RMDIR , FENUM
		return new Value(0);
	}
	
	// ---------------- リクエスト値の取得 ---------------- //
	
	public Value syscall_REQ__period__COMMAND(Vector args) throws Exception {
		// リクエストのコマンド文字列を返します。
		return new Value(getAya().getMethodOfLastRequest());
	}
	
	public Value syscall_REQ__period__PROTOCOL(Vector args) throws Exception {
		// リクエストのプロトコル名称とバージョン部分を返します。
		return new Value(getAya().getProtocolOfLastRequest());
	}
	
	public Value syscall_REQ__period__HEADER(Vector args) throws Exception {
		// n 番目のヘッダのキー名称を取得します。
		String key = getAya().getKeyOfLastRequest((int)content(args.elementAt(0)).getInteger());
		return new Value(key);
	}
	
	public Value syscall_REQ__period__KEY(Vector args) throws Exception {
		// REQ.HEADERと同じ。
		return syscall_REQ__period__HEADER(args);
	}
	
	public Value syscall_REQ__period__VALUE(Vector args) throws Exception {
		/*
			この関数はオーバーロードされており、引数にはキーの名前、もしくは数値を与えることができます。
			引数にキー名称を与えると、その値が取得できます。
			数値の場合は REQ.HEADER 関数と同じく、 n 番目のヘッダ値取得となります。n は 0 origin です。
			いずれの場合でも、イリーガルな引数が指定された場合は空文字列が返ります。
		*/
		Value arg = content(args.get(0));
		if (arg.isInteger()) {
			String value = getAya().getValueOfLastRequest((int)arg.getInteger());
			// どうやら値が数値なら数値を返さなきゃならないらしい。
			try {
				return new Value(Double.parseDouble(value));
			}
			catch (Exception e) {
				return new Value(value);
			}
		}
		else if (arg.isString()) {
			String value = getAya().getValueOfLastRequest(arg.getString());
			try {
				return new Value(Double.parseDouble(value));
			}
			catch (Exception e) {
				return new Value(value);
			}
		}
		return new Value("");
	}
	
	// ---------------- 汎用 DLL 呼び出し ---------------- //
	
	public Value syscall_LOADLIB(Vector args) throws Exception {
		warnUseOfDllFunctions();
		return null;
	}
	
	public Value syscall_UNLOADLIB(Vector args) throws Exception {
		warnUseOfDllFunctions();
		return null;
	}
	
	public Value syscall_REQUESTLIB(Vector args) throws Exception {
		warnUseOfDllFunctions();
		return null;
	}
	
	public Value syscall_LIB__period__STATUSCODE(Vector args) throws Exception {
		warnUseOfDllFunctions();
		return new Value("");
	}
	
	public Value syscall_LIB__period__PROTOCOL(Vector args) throws Exception {
		warnUseOfDllFunctions();
		return new Value("");
	}
	
	public Value syscall_LIB__period__HEADER(Vector args) throws Exception {
		warnUseOfDllFunctions();
		return new Value("");
	}
	
	public Value syscall_LIB__period__KEY(Vector args) throws Exception {
		return syscall_LIB__period__HEADER(args);
	}
	
	public Value syscall_LIB__period__VALUE(Vector args) throws Exception {
		warnUseOfDllFunctions();
		return new Value("");
	}
	
	boolean have_warned_use_of_dll_functions_before = false;
	public void warnUseOfDllFunctions() {
		if (!have_warned_use_of_dll_functions_before) {
			have_warned_use_of_dll_functions_before = true;
			System.err.println("Warning: This ghost of the Aya uses system functions of calling DLL. Java implementation of the Aya doesn't support them.");
		}
	}
	
	// ---------------- ファイル操作 ---------------- //
	
	public Value syscall_FOPEN(Vector args) throws Exception {
		/*
			ファイルをオープンします。読み取りや書き込みを行う前に、必ず対象のファイルをオープンします。
			引数は 2 つで、第一引数は対象のファイル名（aya.dll 位置からの相対パス指定、絶対パス指定のいずれも可能）、第二引数がオープンモードです。
			
			オープンモードは以下の 3 種類より選びます。
			
			モード　　　意味
			───────────────────────────────────────
			"read"　　　読み取り
			"write" 　　書き込み。指定ファイルが既に存在している場合は上書きされる。
			"append"　　書き込み。指定ファイルが既に存在している場合はその終端に追加される。
			
			戻り値　0/1/2 = 失敗/成功/既に指定ファイルはオープンしている
		*/
		warnUseOfFileIOFunctions();
		return new Value(0);
	}
	
	public Value syscall_FCLOSE(Vector args) throws Exception {
		/*
			ファイルをクローズします。読み取りや書き込みを行った後は、必ず対象のファイルをクローズします。
			引数にはクローズするファイル名（aya.dll 位置からの相対パス指定、絶対パス指定のいずれも可能）を指定します。
			
			戻り値はありません。
		*/
		warnUseOfFileIOFunctions();
		return null;
	}
	
	public Value syscall_FREAD(Vector args) throws Exception {
		/*
			"read" でオープンしたファイルから 1 行読み取ります。
			引数にはクローズするファイル名（aya.dll 位置からの相対パス指定、絶対パス指定のいずれも可能）を指定します。
			
			戻り値は読み取った文字列です。読み取った文字列の終端に改行文字はありません。
			EOF（ファイルの最後）まで達したときは数値の -1 が返ります。
			
			例として、test.txt の内容をそのままバルーンに表示するコードを以下に示します。
			test.txt を read でオープン、1 行読み取って \n を付け加えるのを繰り返し、最後まで読み終わったらファイルをクローズ、という流れです。
		*/
		warnUseOfFileIOFunctions();
		return new Value("");
	}
	
	public Value syscall_FWRITE(Vector args) throws Exception {
		/*
			"write"、"append" でオープンしたファイルへ 1 行書き込みます（処理の最後に改行コードを自動的に書き込みます）。
			引数は 2 つで、第一引数は対象のファイル名（aya.dll 位置からの相対パス指定、絶対パス指定のいずれも可能）、第二引数が書き込む内容です。
			
			戻り値はありません。
		*/
		warnUseOfFileIOFunctions();
		return null;
	}
	
	public Value syscall_FWRITE2(Vector args) throws Exception {
		/*
			"write"、"append" でオープンしたファイルへデータを書き込みます。
			FWRITE と異なり、改行コードは書き込みません。
		*/
		warnUseOfFileIOFunctions();
		return null;
	}
	
	public Value syscall_FCOPY(Vector args) throws Exception {
		/*
			ファイルを別のディレクトリにコピーします。
			第一引数にコピー対象のファイル名、第二引数にコピー先のディレクトリを指定します。パス指定は aya.dll 位置からの相対パス指定、絶対パス指定のいずれも可能です。
			戻り値　0/1 = 失敗/成功
		*/
		warnUseOfFileIOFunctions();
		return new Value(0);
	}
	
	public Value syscall_FMOVE(Vector args) throws Exception {
		/*
			ファイルを別のディレクトリに移動します。
			第一引数に移動対象のファイル名、第二引数に移動先のディレクトリを指定します。パス指定は aya.dll 位置からの相対パス指定、絶対パス指定のいずれも可能です。
			戻り値　0/1 = 失敗/成功
		*/
		warnUseOfFileIOFunctions();
		return new Value(0);
	}
	
	public Value syscall_FDELETE(Vector args) throws Exception {
		/*
			ファイルを削除します。
			引数に削除対象のファイル名を指定します（aya.dll 位置からの相対パス指定、絶対パス指定のいずれも可能）。
			戻り値　0/1 = 失敗/成功
		*/
		warnUseOfFileIOFunctions();
		return new Value(0);
	}
	
	public Value syscall_FRENAME(Vector args) throws Exception {
		/*
			ファイルの名前を変更します。
			引数は 2 つで、第一引数が変更前、第二引数が変更後のファイル名です（aya.dll 位置からの相対パス指定、絶対パス指定のいずれも可能）。
			戻り値　0/1 = 失敗/成功
		*/
		warnUseOfFileIOFunctions();
		return new Value(0);
	}
	
	public Value syscall_FSIZE(Vector args) throws Exception {
		/*
			ファイルサイズを単位バイトで取得します。
			引数に対象のファイル名を指定します（aya.dll 位置からの相対パス指定、絶対パス指定のいずれも可能）。
			結果は戻り値で得られます。処理に失敗した場合は -1 が返ります。
		*/
		warnUseOfFileIOFunctions();
		return new Value(-1);
	}
	
	public Value syscall_MKDIR(Vector args) throws Exception {
		/*
			ディレクトリを作成します。
			引数に作成するディレクトリ名を指定します。aya.dll 位置からの相対パス指定、絶対パス指定のいずれも可能です。
			戻り値　0/1 = 失敗/成功
		*/
		warnUseOfFileIOFunctions();
		return new Value(0);
	}
	
	public Value syscall_RMDIR(Vector args) throws Exception {
		/*
			空のディレクトリを削除します。空でないディレクトリは削除できません。
			引数に削除するディレクトリ名を指定します。aya.dll 位置からの相対パス指定、絶対パス指定のいずれも可能です。
			戻り値　0/1 = 失敗/成功
		*/
		warnUseOfFileIOFunctions();
		return new Value(0);
	}
	
	public Value syscall_FENUM(Vector args) throws Exception {
		/*
			指定されたディレクトリ内のファイル/ディレクトリのリストを作成します。
			引数に対象のディレクトリ位置を aya.dll からの相対位置で指定します。aya.dll 直下であれば "" です。パス指定は aya.dll 位置からの相対パス指定、絶対パス指定のいずれも可能です。
			結果は戻り値で得られます。内容は半角カンマで区切られたファイル/ディレクトリ名のリストです。ディレクトリ名の先頭には \ が付されており、これを見ることで通常のファイルと区別できます。
			
			区切り文字は半角カンマから他の文字(列)へ変更することができます。第二引数で指定してください。
		*/
		warnUseOfFileIOFunctions();
		return new Value("");
	}
	
	boolean have_warned_use_of_fileio_functions_before = false;
	public void warnUseOfFileIOFunctions() {
		if (!have_warned_use_of_fileio_functions_before) {
			have_warned_use_of_fileio_functions_before = true;
			System.err.println("Warning: This ghost of the Aya uses system functions of File IO. Java implementation of the Aya doesn't support them yet.");
		}
	}
	
	// ---------------- 文字列操作 ---------------- //
	
	public Value syscall_STRLEN(Vector args) throws Exception {
		/*
			文字列長を取得します。単位はバイトです。
			第一引数に長さを調べる文字列を指定します。結果は戻り値で取得します。
			第二引数に結果を取得する変数を指定することもできます。この場合戻り値はありません。
		*/
		String str = content(args.elementAt(0)).getString();
		Value result = new Value(str.getBytes(Aya.CHARSET).length);
		
		if (args.size() >= 2) {
			VariablePointer vptr = (VariablePointer)args.elementAt(1);
			vptr.store(result);
			return null;
		}
		else {
			return result;
		}
	}
	
	public Value syscall_STRSTR(Vector args) throws Exception {
		/*
			STRSTRは文字列の中から部分文字列を検索し、見つかったバイト位置を返します。
			引数は 3 つ、もしくは 4 つです。第一引数は検索される文字列、第二引数は検索する文字列、第三引数は探し始めるバイト位置です。第四引数は結果を取得する変数ですが、これを省略すると結果を戻り値で取得できます。
			第三引数が被検索対象の文字列長より大きい場合、および部分文字列が見つからなかった場合は -1 が返ります。
		*/
		String all = content(args.get(0)).getString();
		String part = content(args.get(1)).getString();
		int beginning = (int)content(args.get(2)).getInteger();
		
		Value result;
		int pos = all.indexOf(part,beginning);
		if (pos == -1) {
			result = new Value(-1);
		}
		else {
			result = new Value(all.substring(0,pos).getBytes(Aya.CHARSET).length);
		}
		
		if (args.size() >= 4) {
			VariablePointer vptr = (VariablePointer)args.elementAt(3);
			vptr.store(result);
			return null;
		}
		else {
			return result;
		}
	}
	
	public Value syscall_SUBSTR(Vector args) throws Exception {
		/*
			SUBSTRは文字列から部分文字列を抽出します。
			引数は 3 つ。第一引数が抽出元の文字列、第二引数は抽出を開始するバイト位置、第三引数は抽出するバイト数です。
			結果は戻り値として取得します。
		*/
		byte[] source = content(args.get(0)).getString().getBytes(Aya.CHARSET);
		int beginning = (int)content(args.get(1)).getInteger();
		int num_bytes = (int)content(args.get(2)).getInteger();
		
		if (beginning + num_bytes >= source.length || beginning < 0 || num_bytes < 0) {
			return new Value("");
		}
		else {
			ByteArrayOutputStream baos = new ByteArrayOutputStream();
			baos.write(source,beginning,num_bytes);
			return new Value(baos.toString(Aya.CHARSET));
		}
	}
	
	public Value syscall_REPLACE(Vector args) throws Exception {
		/*
			REPLACE は、文字列中に出現する部分文字列のすべてを、別の文字列に置換えます。
			引数は 3 つで第一引数が被置換文字列、第二引数は置換前の部分文字列、第三引数は置換後の部分文字列です。
			結果は戻り値として取得します。
		*/
		String source = content(args.get(0)).getString();
		String before = content(args.get(1)).getString();
		String after = content(args.get(2)).getString();
		return new Value(Pattern.compile("\\Q"+before+"\\E").matcher(source).replaceAll(after));
	}
	
	public Value syscall_ERASE(Vector args) throws Exception {
		/*
			ERASEは文字列の一部を削除します。

			引数は 3 つで、第一引数が元の文字列、第二引数は削除を開始するバイト位置、第三引数は削除するバイト数です。
			結果は戻り値として取得します。
		*/
		byte[] source = content(args.get(0)).getString().getBytes(Aya.CHARSET);
		int beginning = (int)content(args.get(1)).getInteger();
		int num_bytes = (int)content(args.get(2)).getInteger();
		
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		baos.write(source,0,beginning);
		baos.write(source,beginning + num_bytes,source.length - (beginning + num_bytes));
		return new Value(baos.toString(Aya.CHARSET));
	}
	
	public Value syscall_INSERT(Vector args) throws Exception {
		/*
			INSERTは文字列に部分文字列を挿入します。
			引数は 3 つで、第一引数が元の文字列、第二引数は部分文字列を挿入するバイト位置、第三引数は挿入する部分文字列です。
			結果は戻り値として取得します。
			挿入バイト位置が負数の場合は先頭に挿入されます。挿入元文字列長より大きい値を与えた場合は終端に追加されます。
		*/
		byte[] source = content(args.get(0)).getString().getBytes(Aya.CHARSET);
		int beginning = (int)content(args.get(1)).getInteger();
		byte[] part = content(args.get(2)).getString().getBytes(Aya.CHARSET);
		
		if (beginning < 0) {
			beginning = 0;
		}
		else if (beginning > source.length) {
			beginning = source.length;
		}
		
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		baos.write(source,0,beginning);
		baos.write(part,0,part.length);
		baos.write(source,beginning,source.length - beginning);
		return new Value(baos.toString(Aya.CHARSET));
	}
	
	public Value syscall_TOUPPER(Vector args) throws Exception {
		/*
			TOUPPER は文字列に含まれるすべての半角英小文字を半角英大文字に変換します。
			引数に変換元文字列を指定します。変換結果は戻り値として取得します。
		*/
		return new Value(content(args.elementAt(0)).getString().toUpperCase());
	}
	
	public Value syscall_TOLOWER(Vector args) throws Exception {
		/*
			TOLOWER は文字列に含まれるすべての半角英大文字を半角英小文字に変換します。
			引数に変換元文字列を指定します。変換結果は戻り値として取得します。
		*/
		return new Value(content(args.elementAt(0)).getString().toLowerCase());
	}
	
	public Value syscall_CUTSPACE(Vector args) throws Exception {
		/*
			文字列前後の空白文字だけからなる部分を削除します。
			該当する空白文字は半角スペース、全角スペース、タブ文字です。
			引数に変換元文字列を指定します。変換結果は戻り値として取得します。
		*/
		return new Value(content(args.elementAt(0)).getString().trim());
	}
	
	public Value syscall_MSTRLEN(Vector args) throws Exception {
		/*
			文字列の文字数を取得します。
			第一引数に長さを調べる文字列を指定します。結果は戻り値で取得します。
		*/
		return new Value(content(args.get(0)).getString().length());
	}
	
	public Value syscall_MSTRSTR(Vector args) throws Exception {
		/*
			文字列の中から部分文字列を検索し、見つかった文字位置を返します。
			引数は 3 つです。第一引数は検索される文字列、第二引数は検索する文字列、第三引数は探し始める文字位置です。
		*/
		String part = content(args.get(0)).getString();
		String all = content(args.get(1)).getString();
		int beginning = (int)content(args.get(2)).getInteger();
		
		return new Value(all.indexOf(part,beginning));
	}
	
	public Value syscall_MSUBSTR(Vector args) throws Exception {
		/*
			SUBSTRは文字列から部分文字列を文字単位で抽出します。
			引数は 3 つ。第一引数が抽出元の文字列、第二引数は抽出を開始する文字位置、第三引数は抽出する文字数です。
			結果は戻り値として取得します。
		*/
		String source = content(args.get(0)).getString();
		int beginning = (int)content(args.get(1)).getInteger();
		int num_bytes = (int)content(args.get(2)).getInteger();
		
		return new Value(source.substring(beginning,beginning + num_bytes));
	}
	
	public Value syscall_MERASE(Vector args) throws Exception {
		/*
			ERASEは文字列の一部を削除します。
			引数は 3 つで、第一引数が元の文字列、第二引数は削除を開始する文字位置、第三引数は削除する文字数です。
			結果は戻り値として取得します。
		*/
		String source = content(args.get(0)).getString();
		int beginning = (int)content(args.get(1)).getInteger();
		int num_bytes = (int)content(args.get(2)).getInteger();
		
		StringBuffer buf = new StringBuffer(source);
		buf.delete(beginning,num_bytes);
		return new Value(buf.toString());
	}
	
	public Value syscall_MINSERT(Vector args) throws Exception {
		/*
			INSERTは文字列に部分文字列を挿入します。
			引数は 3 つで、第一引数が元の文字列、第二引数は部分文字列を挿入する文字位置、第三引数は挿入する部分文字列です。
			結果は戻り値として取得します。
		*/
		String source = content(args.get(0)).getString();
		int beginning = (int)content(args.get(1)).getInteger();
		String part = content(args.get(2)).getString();
		
		if (beginning < 0) {
			beginning = 0;
		}
		else if (beginning > source.length()) {
			beginning = source.length();
		}
		
		StringBuffer buf = new StringBuffer(source);
		buf.insert(beginning,part);
		return new Value(buf.toString());
	}
	
	// ---------------- 算術関数 ---------------- //
	
	public Value syscall_SIN(Vector args) throws Exception {
		/*
			sin を計算します。引数の単位はラジアンです。
			結果は戻り値で取得します。
		*/
		double rad = content(args.get(0)).getReal();
		return new Value(Math.sin(rad));
	}
	
	public Value syscall_COS(Vector args) throws Exception {
		/*
			cos を計算します。引数の単位はラジアンです。
			結果は戻り値で取得します。
		*/
		double rad = content(args.get(0)).getReal();
		return new Value(Math.cos(rad));
	}
	
	public Value syscall_TAN(Vector args) throws Exception {
		/*
			tan を計算します。引数の単位はラジアンです。
			結果は戻り値で取得します。
		*/
		double rad = content(args.get(0)).getReal();
		return new Value(Math.tan(rad));
	}
	
	public Value syscall_LOG(Vector args) throws Exception {
		/*
			e を底とする対数（自然対数）を計算します。結果は戻り値で取得します。
		*/
		double n = content(args.get(0)).getReal();
		return new Value(Math.log(n));
	}
	
	public Value syscall_LOG10(Vector args) throws Exception {
		/*
			10 を底とする対数（常用対数）を計算します。結果は戻り値で取得します。
		*/
		double n = content(args.get(0)).getReal();
		return new Value(Math.log(n) / Math.log(10));
	}
	
	public Value syscall_POW(Vector args) throws Exception {
		/*
			第一引数の第二引数乗を計算します。結果は戻り値で取得します。
		*/
		double a = content(args.get(0)).getReal();
		double b = content(args.get(1)).getReal();
		return new Value(Math.pow(a,b));
	}
	
	public Value syscall_FLOOR(Vector args) throws Exception {
		double n = content(args.get(0)).getReal();
		return new Value(Math.floor(n));
	}
	
	public Value syscall_CEIL(Vector args) throws Exception {
		double n = content(args.get(0)).getReal();
		return new Value(Math.ceil(n));
	}
	
	public Value syscall_ROUND(Vector args) throws Exception {
		double n = content(args.get(0)).getReal();
		return new Value(Math.round(n));
	}
	
	public Value syscall_SQRT(Vector args) throws Exception {
		/*
			平方根を計算します。結果は戻り値で取得します。
			引数に負数は与えられません。イリーガルコードとして -1 を返します。
		*/
		double n = content(args.get(0)).getReal();
		if (n < 0) {
			return new Value(-1);
		}
		else {
			return new Value(Math.sqrt(n));
		}
	}
	
	public Value content(Object obj) {
		// objがValueならそのまま、VariablePointerならデリファレンスして返す。
		if (obj instanceof Value) {
			return (Value)obj;
		}
		else if (obj instanceof VariablePointer) {
			return ((VariablePointer)obj).fetch();
		}
		else {
			throw new RuntimeException("Internal Error: an object of "+obj.getClass().getName()+" was in arguments of system call illegally.");
		}
	}
}
