/*
 "NiseRingo" is a clone of "Niseharuna" works on MacOS X.
 Copyright (C) 2001-2003 phonohawk

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
package aya.node;
import aya.*;
import aya.node.*;
import aya.eval.*;
import java.util.*;

/*
	case文は非常に複雑であるため、Statementの内部クラスではなく独立したファイルで定義します。
	Statementのサブクラスです。
*/

public class Case extends Statement
{
	// case文は、一つの比較対象となる式と、一つ以上のcase候補から成る。
	
	Expression attention; // 比較対象となる式
	Vector candidates; // [Candidate]
	
	protected static Case _factory = null;
	public static Case getCaseFactory() {
		if (_factory == null) {
			_factory = new Case();
		}
		return _factory;
	}
	protected Case() {
		attention = null;
		candidates = null;
	}
	
	public Case(Expression attention)
	{
		this.attention = attention;
		this.candidates = new Vector();
	}
	
	public Case addCandidate(Candidate cand) {
		candidates.add(cand);
		return this;
	}
	
	public Object eval(Namespace ns) throws Exception {
		Value val = attention.eval(ns);
		
		// まずはothers以外で候補を探す。
		for (int i = 0; i < candidates.size(); i++) {
			Candidate cand = (Candidate)candidates.elementAt(i);
			
			if (!cand.isOthers() && cand.match(val)) {
				// マッチするなら、このブロックを評価して返す。
				return cand.eval(ns);
			}
		}
		
		// othersを見る。
		for (int i = 0; i < candidates.size(); i++) {
			Candidate cand = (Candidate)candidates.elementAt(i);
			
			if (cand.isOthers()) {
				return cand.eval(ns);
			}
		}
		
		return null;
	}
	
	public Candidate newCandidate(CaseCondition condition,Block block) {
		return new Candidate(condition,block);
	}
	public class Candidate {
		// case候補には、when候補とothers候補の二つがある。
		// when候補はcase条件とブロックを、others候補はブロックのみを持つ。
		CaseCondition condition; // nullであれば、others候補。
		Block block;
		
		public Candidate(CaseCondition condition,Block block) {
			this.condition = condition;
			this.block = block;
		}
		
		public boolean isOthers() {
			return condition == null;
		}
		
		public boolean match(Value attention) {
			if (condition == null) {
				return true;
			}
			else {
				return condition.match(attention);
			}
		}
		
		public Value eval(Namespace ns) throws Exception {
			return block.eval(new Namespace(ns));
		}
	}
	
	public CaseCondition newCondition() {
		return new CaseCondition();
	}
	public class CaseCondition {
		// case条件は一つ以上のcase部分条件から成る。
		Vector subconditions; // [CaseSubcondition]
		
		public CaseCondition() {
			this.subconditions = new Vector();
		}
		
		public CaseCondition addSubcondition(CaseSubcondition sub) {
			subconditions.add(sub);
			return this;
		}
		
		public boolean match(Value attention) {
			// case部分条件が複数あれば、どれか一つでも一致すれば真を返す。
			for (int i = 0; i < subconditions.size(); i++) {
				CaseSubcondition sub = (CaseSubcondition)subconditions.elementAt(i);
				
				if (sub.match(attention)) {
					return true;
				}
			}
			return false;
		}
	}
	
	public CaseSubcondition newSubcondition(double num) {
		return new SubcondNumeric(num);
	}
	public CaseSubcondition newSubcondition(double lower,double higher) {
		return new SubcondNumericRange(lower,higher);
	}
	public CaseSubcondition newSubcondition(String str) {
		return new SubcondString(str);
	}
	public CaseSubcondition newSubcondition(String lower,String higher) {
		return new SubcondStringRange(lower,higher);
	}
	public class CaseSubcondition {
		// case部分条件には、数値範囲、文字列範囲、または式の3通りの形式がある。
		
		public boolean match(Value attention) {
			try {
				throw new Exception("Abstract method \"eval\" of CaseSubcondition has called directory.");
			}
			catch (Exception e) {return false;}
		}
	}
	public class SubcondNumeric extends CaseSubcondition {
		double num;
		
		public SubcondNumeric(double num) {
			this.num = num;
		}
		
		public boolean match(Value attention) {
			if (attention.isNumeric()) {
				return num == attention.getReal();
			}
			else {
				return false;
			}
		}
	}
	public class SubcondNumericRange extends CaseSubcondition {
		double lower,higher;
		
		public SubcondNumericRange(double lower,double higher) {
			this.lower = lower;
			this.higher = higher;
		}
		
		public boolean match(Value attention) {
			if (attention.isNumeric()) {
				double n = attention.getReal();
				return (lower <= n && n <= higher) || (higher <= n && n <= lower);
			}
			else {
				return false;
			}
		}
	}
	public class SubcondString extends CaseSubcondition {
		String str;
		
		public SubcondString(String str) {
			this.str = str;
		}
		
		public boolean match(Value attention) {
			if (attention.isString()) {
				return attention.getString().equals(str);
			}
			else {
				return false;
			}
		}
	}
	public class SubcondStringRange extends CaseSubcondition {
		String lower,higher;
		
		public SubcondStringRange(String lower,String higher) {
			this.lower = lower;
			this.higher = higher;
		}
		
		public boolean match(Value attention) {
			if (attention.isString()) {
				String str = attention.getString();
				return (lower.compareTo(str) <= 0 && str.compareTo(higher) <= 0) ||
					   (higher.compareTo(str) <= 0 && str.compareTo(lower) <= 0);
			}
			else {
				return false;
			}
		}
	}
}
