/*
 "NiseRingo" is a clone of "Niseharuna" works on MacOS X.
 Copyright (C) 2001-2003 phonohawk

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
package aya.node;
import aya.*;
import aya.node.*;
import aya.eval.*;
import java.util.*;

public class Statement
{
	protected static Statement _factory = null;
	
	public static Statement getFactory() {
		if (_factory == null) {
			_factory = new Statement();
		}
		return _factory;
	}
	
	protected Statement() {
		// for factory
	}
	
	public Object eval(Namespace ns) throws Exception {
		// Statementのeval()は、null、ValueまたはVector<Value>を返す。
		throw new Exception("Abstract method \"eval\" of Statement has called directory.");
	}
	
	public Statement newStatement(Expression expr) {
		return new ExpressionalStatement(expr);
	}
	
	public Statement newStatement(Substitution subs) {
		return new SubstitutionalStatement(subs);
	}
	
	public Statement newStatement(Block block) {
		return new BlockStatement(block);
	}
	
	public If newIf(Condition cond,Block block) {
		return new If(cond,block);
	}
	
	public Switch newSwitch(Expression expr,Block block) {
		return new Switch(expr,block);
	}
	
	public While newWhile(Condition cond,Block block) {
		return new While(cond,block);
	}
	
	public For newFor(Substitution init,Condition cond,Substitution alter,Block block) {	
		return new For(init,cond,alter,block);
	}
	
	public Separate newSeparate() {
		return new Separate();
	}
	
	public Break newBreak() {
		return new Break();
	}
	
	public Continue newContinue() {
		return new Continue();
	}
	
	public Return newReturn() {
		return new Return();
	}
	
	public class ExpressionalStatement extends Statement {
		Expression expr;
		
		public ExpressionalStatement(Expression expr) {
			this.expr = expr;
		}
		
		public Object eval(Namespace ns) throws Exception {
			return expr.eval(ns);
		}
	}
	public class SubstitutionalStatement extends Statement {
		Substitution subs;
		
		public SubstitutionalStatement(Substitution subs) {
			this.subs = subs;
		}
		
		public Object eval(Namespace ns) throws Exception {
			subs.eval(ns);
			return null;
		}
	}
	public class BlockStatement extends Statement {
		Block block;
		
		public BlockStatement(Block block) {
			this.block = block;
		}
		
		public Object eval(Namespace ns) throws Exception {
			return block.eval(new Namespace(ns));
		}
	}
	public class If extends Statement {
		Condition cond; // ifの条件
		Block block_if; // ifのブロック
		
		Vector elseif_conds; // [(Condition)それぞれのelseifの条件] 一つも無ければnull
		Vector elseif_blocks; // [(Block)それぞれのelseifのブロック] 一つも無ければnull
		
		Block block_else; // elseのブロック。無ければnull
		
		public If(Condition cond,Block block_if) {
			this.cond = cond;
			this.block_if = block_if;
			this.elseif_conds = null;
			this.elseif_blocks = null;
			this.block_else = null;
		}
		
		public If addElseIf(Condition cond,Block block) {
			if (elseif_conds == null) {
				elseif_conds = new Vector();
				elseif_blocks = new Vector();
			}
			elseif_conds.add(cond);
			elseif_blocks.add(block);
			return this;
		}
		
		public If setElse(Block block) {
			block_else = block;
			return this;
		}
		
		public Object eval(Namespace ns) throws Exception {
			if (cond.eval(ns)) {
				return block_if.eval(new Namespace(ns));
			}
			
			if (elseif_conds != null) {
				for (int i = 0; i < elseif_conds.size(); i++) {
					Condition elseif_cond = (Condition)elseif_conds.elementAt(i);
					Block elseif_block = (Block)elseif_blocks.elementAt(i);
					
					if (elseif_cond.eval(ns)) {
						return elseif_block.eval(new Namespace(ns));
					}
				}
			}
			
			if (block_else != null) {
				return block_else.eval(new Namespace(ns));
			}
			
			return null;
		}
	}
	public class Switch extends Statement {
		Expression expr;
		Block block;
		
		public Switch(Expression expr,Block block) {
			this.expr = expr;
			this.block = block;
		}
		
		public Object eval(Namespace ns) throws Exception {
			return block.eval(new Namespace(ns),(int)expr.eval(ns).getInteger());
		}
	}
	public class While extends Statement {
		Condition cond;
		Block block;
		
		public While(Condition cond,Block block) {
			this.cond = cond;
			this.block = block;
		}
		
		public Object eval(Namespace ns) throws Exception {
			Vector result = null;
			
			Namespace ns_for_loop = new Namespace(ns);
			while (cond.eval(ns_for_loop)) {
				try {
					Value val = block.eval(ns_for_loop);
					if (val != null) {
						if (result == null) {
							result = new Vector();
						}
						result.add(val);
					}
				}
				catch (BreakOccurrence b) {
					// ここで終わり。
					break;
				}
				catch (ContinueOccurrence c) {
					// 特に何もせず、単にcatch。
				}
			}
			
			return result;
		}
	}
	public class For extends Statement {
		Substitution init;
		Condition cond;
		Substitution alter;
		Block block;
		
		public For(Substitution init,Condition cond,Substitution alter,Block block) {
			// block以外は省略可能。
			this.init = init;
			this.cond = cond;
			this.alter = alter;
			this.block = block;
		}
		
		public Object eval(Namespace ns) throws Exception {
			Vector result = null;
			
			Namespace ns_for_loop = new Namespace(ns);
			for (init.eval(ns_for_loop); cond.eval(ns_for_loop); alter.eval(ns_for_loop)) {
				try {
					Value val = block.eval(ns_for_loop);
					if (val != null) {
						if (result == null) {
							result = new Vector();
						}
						result.add(val);
					}
				}
				catch (BreakOccurrence b) {
					// ここで終わり。
					break;
				}
				catch (ContinueOccurrence c) {
					// 特に何もせず、単にcatch。
				}
			}
			
			return result;
		}
	}
	public class Separate extends Statement {}
	public class Break extends Statement {}
	public class Continue extends Statement {}
	public class Return extends Statement {}
	public boolean isSeparator() {
		return (this instanceof Separate);
	}
	public boolean isBreak() {
		return (this instanceof Break);
	}
	public boolean isContinue() {
		return (this instanceof Continue);
	}
	public boolean isReturn() {
		return (this instanceof Return);
	}
	
	public BreakOccurrence newBreakOccurrence() {
		return new BreakOccurrence();
	}
	public class BreakOccurrence extends Exception {}
	
	public ContinueOccurrence newContinueOccurrence() {
		return new ContinueOccurrence();
	}
	public class ContinueOccurrence extends Exception {}
}
